/*	$ DOOROOS: drv_pwm.c v 0.1 11-12-03 21:47 $	*/

/*
 * Copyright (C) 2015	iBOSN systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are not permitted completely.
 *
 */

/*
 * Module Name: drv_pwm.c
 *
 * Purpose: Start of JBOSN Operating System.
 *          make the utility for wave device driver
 *
 */
#include <types.h>
#include <thread.h>
#include <message.h>
#include <interrupt.h>
#include <device.h>
#include <error.h>

#include <time.h>

#include <panic.h>

#include <mach/hal_interrupt.h>
#include <mach/SAMA5D3X.h>
#include <mach/SAMA5D3Xbd.h>
#include <mach/gpio.h>

#include <lib/string.h>
#include <dev/pwmdrv.h>

//==========================================================================
STATIC BOOL                     th_PwmDevice( UINT32 Arg );

#define PWMDEV_STACK_SIZE      (512)
STATIC UINT32                   PwmDevStack[PWMDEV_STACK_SIZE>>2];

STATIC DEVICECONFIG             PwmDeviceConfig ={"PWM", 0, -1, NULL, NULL };

PUBLIC THREADCONFIG             PwmDev_Thread =
                                            {
                                                (UINT32)&PwmDeviceConfig,
                                                th_PwmDevice,
                                                (PBYTE)PwmDevStack,
                                                PWMDEV_STACK_SIZE,
                                                THREAD_PRIORITY_DRIVER+10
                                            };
//==========================================================================
typedef struct _PWM_INFO_
{
    INT32       OnDuration;
    BOOL        bOnOff;
}PWM_INFO, *PWM_PINFO;

STATIC PWM_INFO     PwmInfo[2];

STATIC UINT32       Channel;

#define PWM_CHANNEL                 2
#define PWM_DEFAULT_FREQUENCY       (10000)

//==========================================================================
static void PWMC_ConfigureClocks(AT91S_PWMC* pPwm, UINT32 mode)
{
    pPwm->PWM_CLK= mode;
}

static void PWMC_EnableChannel(AT91S_PWMC* pPwm,UINT32 channel)
{
    pPwm->PWM_ENA= 0x1ul<<channel;  
}
static void PWMC_DisableChannel(AT91S_PWMC* pPwm,UINT32 channel)
{
    pPwm->PWM_DIS= 0x1ul<<channel;    
}

void PWMC_ConfigureChannel(AT91S_PWMC* pPwm,UINT32 channel,UINT32 mode)
{
    pPwm->PWM_CH_NUM[channel].PWM_CMR= mode;
}

static void PWMC_SetPeriod( AT91S_PWMC* pPwm, UINT32 channel, UINT32 period)
{
    /* If channel is disabled, write to CPRD */
    if ((pPwm->PWM_SR & (1 << channel)) == 0)
    {
        pPwm->PWM_CH_NUM[channel].PWM_CPRD = period;
    }
    /* Otherwise use update register */
    else
    {
        pPwm->PWM_CH_NUM[channel].PWM_CPRDUPD = period;
    }
}

static void PWMC_SetDutyCycle( AT91S_PWMC* pPwm, UINT32 channel, UINT32 duty)
{
    /* If channel is disabled, write to CDTY */
    if ((pPwm->PWM_SR & (1 << channel)) == 0)
    {
        pPwm->PWM_CH_NUM[channel].PWM_CDTY = duty;
    }
    /* Otherwise use update register */
    else
    {
        pPwm->PWM_CH_NUM[channel].PWM_CDTYUPD = duty;
    }
}

#define PWM_OUT_CHANNEL0    2
#define PWM_OUT_CHANNEL1    3

STATIC VOID
PwmInit( VOID )
{
//DBGLINE(1,"\n");
// Configure PIO controllers to periph mode
    const struct pio_desc PWM_pio[] = 
    {
        {"PWM3L",     AT91C_PIN_PB(13),   0, PIO_DEFAULT, PIO_OUTPUT_1},
        {"PWM2L",     AT91C_PIN_PD(6),    0, PIO_DEFAULT, PIO_OUTPUT_1},
    };
    
    /* PIO configuration */
    //PIO_Configure( &gPinTIOA0, PIO_LISTSIZE( gPinTIOA0 ) ) ;
    pio_setup(PWM_pio, sizeof(PWM_pio)/sizeof(struct pio_desc));
//DBGLINE(1," BOARD_MCK = %d\n", BOARD_MCK);
//DBGLINE(1," SAMA5D3_ID_PWM = %x\n", SAMA5D3_ID_PWM);
    pio_set_value( AT91C_PIN_PB(13), 1 );
    pio_set_value( AT91C_PIN_PD(6), 1 );

    PMC_EnablePeripheral( SAMA5D3_ID_PWM ) ;

    UINT32 mode;
    AT91S_PWMC *pPwm = AT91C_BASE_PWMC;

    /* Set clock A and clock B */
    mode = PWM_CLK_PREB(0x01) | (PWM_CLK_DIVB(6)) | 
           PWM_CLK_PREA(0x01) | (PWM_CLK_DIVA(6));
    PWMC_ConfigureClocks(pPwm, mode);

    pPwm->PWM_SCM = /*(1<<PWM_OUT_CHANNEL0)|(1<<PWM_OUT_CHANNEL1)|*/PWM_SCM_UPDM_MODE1; // mode register
	return;
}


STATIC VOID
PwmSet( UINT32 nChannel )
{
    UINT32 OnDuration = PwmInfo[nChannel].OnDuration;
    AT91S_PWMC *pPwm = AT91C_BASE_PWMC;

DBGLINE(1,"OnDuration = %d\n", OnDuration);
//DBGLINE(1,"pPwm->PWM_CH_NUM[nChannel].PWM_CDTY = %d\n", pPwm->PWM_CH_NUM[nChannel].PWM_CDTY);

    if( OnDuration == 0 )
    {
        if( nChannel == 0 )
        {
            const struct pio_desc PWM_pio[] = 
            {
                {"PWM2L",     AT91C_PIN_PD(6),    0, PIO_DEFAULT, PIO_OUTPUT_1},
            };

            /* PIO configuration */
            pio_setup(PWM_pio, sizeof(PWM_pio)/sizeof(struct pio_desc));
            pio_set_value( AT91C_PIN_PD(6), 1 );

            PWMC_SetDutyCycle(pPwm, PWM_OUT_CHANNEL0, OnDuration);
            PWMC_DisableChannel(pPwm, PWM_OUT_CHANNEL0);
        }
        else
        {
            const struct pio_desc PWM_pio[] = 
            {
                {"PWM3L",     AT91C_PIN_PB(13),   0, PIO_DEFAULT, PIO_OUTPUT_1},
            };
    
            /* PIO configuration */
            pio_setup(PWM_pio, sizeof(PWM_pio)/sizeof(struct pio_desc));
            pio_set_value( AT91C_PIN_PB(13), 1 );
            PWMC_SetDutyCycle(pPwm, PWM_OUT_CHANNEL1, OnDuration);
            PWMC_DisableChannel(pPwm, PWM_OUT_CHANNEL1);
        }

        return;
    }

    
    /* Start the Timer */
    if( nChannel == 0 )
    {

        UINT32 mode;
        AT91S_PWMC *pPwm = AT91C_BASE_PWMC;

        /* If channel is disabled, write to CPRD */
        if ((pPwm->PWM_SR & (1 << PWM_OUT_CHANNEL0)) == 0)
        {
            /* Configure PWM channel0   */
            PWMC_DisableChannel(pPwm, PWM_OUT_CHANNEL0);
            mode=PWM_CMR_CPRE_CLKA|PWM_CMR_DTE | PWM_CMR_CPOL;
            PWMC_ConfigureChannel(pPwm, PWM_OUT_CHANNEL0, mode);
            PWMC_SetPeriod(pPwm, PWM_OUT_CHANNEL0, 1024);
            PWMC_SetDutyCycle(pPwm, PWM_OUT_CHANNEL0, OnDuration);
            PWMC_EnableChannel(pPwm, PWM_OUT_CHANNEL0);

            const struct pio_desc PWM_pio[] = 
            {
                {"PWM2L",     AT91C_PIN_PD(6),    0, PIO_DEFAULT, PIO_PERIPH_C},
            };

            /* PIO configuration */
            pio_setup(PWM_pio, sizeof(PWM_pio)/sizeof(struct pio_desc));
        }
        else
        {
            // channel is enabled
            // update the channel duty cycle
//DBGLINE(1,"OnDuration = %d\n", OnDuration);
            PWMC_SetDutyCycle(pPwm, PWM_OUT_CHANNEL0, OnDuration);
        }
        
    }
    else
    {

        UINT32 mode;
        AT91S_PWMC *pPwm = AT91C_BASE_PWMC;

        /* If channel is disabled, write to CPRD */
        if ((pPwm->PWM_SR & (1 << PWM_OUT_CHANNEL1)) == 0)
        {
            /* Configure PWM channel1 for LED1 */
            PWMC_DisableChannel(pPwm, PWM_OUT_CHANNEL1);
            mode =  PWM_CMR_CPRE_CLKA|PWM_CMR_DTE | PWM_CMR_CPOL; //| PWM_CMR_CALG | PWM_CMR_CPRE_CLKB;
            PWMC_ConfigureChannel(pPwm, PWM_OUT_CHANNEL1, mode);
            PWMC_SetPeriod(pPwm, PWM_OUT_CHANNEL1, 1024);
            PWMC_SetDutyCycle(pPwm, PWM_OUT_CHANNEL1, OnDuration);
            PWMC_EnableChannel(pPwm, PWM_OUT_CHANNEL1);

            const struct pio_desc PWM_pio[] = 
            {
                {"PWM3L",     AT91C_PIN_PB(13),   0, PIO_DEFAULT, PIO_PERIPH_B},
            };
            
            /* PIO configuration */
            pio_setup(PWM_pio, sizeof(PWM_pio)/sizeof(struct pio_desc));
        }
        else
        {
            // channel is enabled
            // update the channel duty cycle
//DBGLINE(1,"OnDuration = %d\n", OnDuration);
            PWMC_SetDutyCycle(pPwm, PWM_OUT_CHANNEL1, OnDuration);
        }
    }

	return;
}


STATIC BOOL
th_PwmDevice( UINT32 Arg )
{
    JB_RECVMESSAGE      Message;
    HANDLE              hPort;
    HANDLE              Handle = NULL;
    PDEVICECONFIG       pDeviceConfig;

    
    pDeviceConfig = (PDEVICECONFIG)Arg;
    
    hPort = PortGet( );    
    PortSetType( hPort, PORTTYPE_MESSAGE|PORTTYPE_INTERRUPT|PORTTYPE_SIGNAL );

    Channel = ChannelGet( hPort );

    pDeviceConfig->Name[3] = (UINT32)pDeviceConfig->Minor + '1';
    pDeviceConfig->Channel = Channel;
    DeviceRegister( (PDEVICECONFIG)pDeviceConfig );

    //InterruptRegister(pDeviceConfig->nInterrupt /*SYSINTR_WAV*/, hPort);
    PwmInit( );
    

	while(1)
	{
		MessageReceive( hPort, &Message);

		switch(Message.Type)
		{
		case MESSAGE_TYPE_SIGNAL:
		{
            continue;
		}

		case MESSAGE_TYPE_INTERRUPT:
            continue;

		case MESSAGE_TYPE_DEVICE_READ:
        {
		    Message.Type = 0;
			break;
	    }

		case MESSAGE_TYPE_DEVICE_WRITE:
	    {
		    Message.Type = 0;
			break;
	    }

		case MESSAGE_TYPE_DEVICE_CONTROL:
		{
            PIOCONTROL_PARAM    pIoControlParam;

            pIoControlParam = (PIOCONTROL_PARAM)Message.Param;
            switch(pIoControlParam->Cmd)
            {
            case IOCTRL_PWM_START_A:
            case IOCTRL_PWM_START_B:
            {
                UINT32 nChannel = pIoControlParam->Cmd - IOCTRL_PWM_START_A;
                PwmInfo[nChannel].OnDuration = (UINT32)(pIoControlParam->pInBuf);
                PwmSet(nChannel);

                Message.Type = TRUE;
                break;
            }

            case IOCTRL_PWM_STOP_A:
            case IOCTRL_PWM_STOP_B:
            {
                UINT32 nChannel = pIoControlParam->Cmd - IOCTRL_PWM_STOP_A;
                PwmInfo[nChannel].OnDuration = 0;
                PwmSet(nChannel);

                Message.Type = TRUE;
                break;
            }

            default:
                Message.Type = FALSE;
                break;
            }

            break;
        }

		case MESSAGE_TYPE_DEVICE_OPEN:
        {
            //Handle = ChannelOpenHandleGetFrom( Channel );
            Handle = Channel;
            Message.Type = Handle;

            break;
        }

		case MESSAGE_TYPE_DEVICE_CLOSE:
        {
		    Handle = NULL;
		    Message.Type = TRUE;
			break;
	    }

        default:
            break;
		}

        MessageReply(&Message);

	}

	return TRUE;	
}




#include <moduleloader.h>

PUBLIC HANDLE 
DllMain( HANDLE hDll, UINT32 Reason, PVOID Reserved )
{
    HANDLE hThread = NULL;
    switch( Reason )
    {
    case DLL_PROCESS_ATTACH:
        hThread = ThreadCreate (&PwmDev_Thread);
    	break;
    
    case DLL_PROCESS_DETACH:
    	break;
    }

    return hThread;
}

